import React, {useState, useEffect, useMemo, useRef}  from 'react';
import logo from './logo.svg';
import './App.css';

import { GraphQLClient } from 'graphql-request';

import createModels2 from './front-models'

let gql = new GraphQLClient("/graphql", {headers: localStorage.authToken ? {Authorization: 'Bearer '+localStorage.authToken} : {}})


//TODO: use graphql-tag to get types, or, at least detect query 
//for proper binding between model objects and result of gql query



const dataReader = async () => {
    const { PriceItem } = await createModels2(gql)
    let data            = await PriceItem.find({}, {limit: [10]})
    return {data, PriceItem }
}

const Th = p =>
<div className='Th' {...p} />

//<div className='Row' style={{position: 'relative',top}}>
const Row = ({top, selected, children}) => { 
    return (
        <>
            <div className={`Row ${selected ? 'selected' : ''}` } >
                {children}
            </div>
        </>
    )
}

const Cell = ({children, options, record, field, selected, ...props}) => {
    let Formatter = React.Fragment

    const viewOrEdit = (!record.constructor.inputs || record.constructor.inputs.some(input => input.name === field.name)) && selected ? 'edit' : 'view'
    if (field.type.name in options[viewOrEdit].formatters){
        Formatter = options[viewOrEdit].formatters[field.type.name]
    }
    if (field.name in options[viewOrEdit].fields){
        Formatter = options[viewOrEdit].fields[field.name]
    }
    if (children && typeof children === 'object'){ 
        if (children.constructor.name in options[viewOrEdit].formatters)
            Formatter = options[viewOrEdit].formatters[children.constructor.name]
        else
            Formatter = options[viewOrEdit].formatters.Object
    }
    return( 
        <div className='Cell' {...props}>
            <Formatter {...props} field={field}>{Formatter === React.Fragment ? (children && children.toString()) : children}</Formatter>
        </div>
    )
}


const ModelListItem = "div"

const ModelList = ({models, selected, onChange, Item=ModelListItem}) => {
    return (
        <aside>
            {Object.entries(models).map(([name, model]) => {
                return (
                    <Item key={name} 
                                   selected={name === selected}
                                   className={name === selected && "selected"}
                                   onClick = {() => onChange(name)}
                                   >
                        {name}
                    </Item>
                )
            })}
        </aside>
    )
}

const Search = ({...props}) => 
<input placeholder="Search..." {...props} className="Search"/>

const Count = ({...props}) => <div className="Count" {...props} /> 

const searchQueryBuilder = (search, model) => {
    const queryRegexp = `/${search.trim().split(/\s+/).join('|')}/`
    return {$or: model.fields.filter(field => field.type.kind == 'SCALAR').map(field => ({[field.name]: queryRegexp}))}
}

const GridHeaderItem = ({field, sort:[sort], ...props}) => 
<div className="GridHeaderItem" {...props}>
    {field.name in sort && sort[field.name] === -1 && '^ '}
    {field.name}
    {field.name in sort && sort[field.name] === 1 && ' v'}
</div>



const GridHeader = ({fields, sort, onSort}) => 
<div className="GridHeader">
    {fields.map(field => <GridHeaderItem field={field} sort={sort} onClick={() => onSort(field.name)}/>)}
</div>

const VirtualScroll = ({options, gridHeight, count, rowHeight, components:Components={Row, Cell}, onScroll, records, skip}) => { 
    //const [records, setRecords] = useState([])
    const limit = gridHeight/rowHeight
    const {Row, Cell} = Components

    const [edit, setEdit] = useState({field: null, record: null})

    const timeout = useRef(0)

    return (
    <div className='GridViewport' 
         onScroll={e => {
             //let skip = Math.ceil(e.target.scrollTop/rowHeight)
             //if (skip < 0) skip = 0
             //if (skip > count - limit) skip = count - limit

             //clearInterval(timeout.current)
             //timeout.current = setTimeout(() => {
                 //console.log(skip, limit)
                 //onScroll(skip,limit)
             //}, 1000)
         }} 
         style={{maxHeight: gridHeight, height: gridHeight}} >
        <div className='GridContent' 
             style={{height: count*rowHeight, minHeight: count*rowHeight, maxHeight: count*rowHeight}} >
        { /*      <div style={{height: skip*rowHeight}} /> */ }
            {records && records.map((record,i) =><React.Fragment key={i}> 
                                            <Row options={options} selected={edit.record === record}>
                                                {record.constructor.fields.map(field => 
                                                    <Cell record={record} key={field.name} field={field} options={options} onClick={() => setEdit({record, field})}
                                                            selected={edit.record === record} >
                                                        {record[field.name]}
                                                    </Cell>)}
                                            </Row>
                {edit.record === record && <Row selected>additional inputs and save/cancel</Row>}
                </React.Fragment>
            )}
        {/*            <div style={{height: (count - skip - records.length)*rowHeight}} /> */ }
        </div>
    </div>
    )
}




const ModelView = ({model, options, components:Components={Search, Count, GridHeader, Grid:VirtualScroll}, rowHeight=150, gridHeight=500, overload=2}) => {



    const [records, setRecords] = useState([])


    const [search, setSearch] = useState("")
    const [count, setCount]   = useState(0)

    const [query,  setQuery]  = useState({})
    const [cursorCalls,  setCursorCalls]  = useState({sort:[{}], skip: [0]/*, limit: [gridHeight/rowHeight]*/})

    const skip = cursorCalls.skip[0]

    console.log(cursorCalls)

    //useEffect(() => {
        //setCursorCalls({sort:cursorCalls.sort, skip: skip ? [skip] : undefined, limit: overloadedRowCount ? [overloadedRowCount] : undefined})
    //}, [scroll])


    useEffect(() => {
        model.count(query, cursorCalls).then(count => setCount(count))
        model.find(query, cursorCalls).then(records => Promise.all(records)).then(records => setRecords(records))
    }, [query, model, cursorCalls])

    const timeout = useRef(0)
    useEffect(() => {
        clearInterval(timeout.current)
        if (!search) 
            setQuery(searchQueryBuilder(search, model))
        else {
            timeout.current = setTimeout(() => {
                setQuery(searchQueryBuilder(search, model))
            },1000)
        }
    },[search, model])

    return (
        <>
                <Components.Search options={options} value={search} onChange={({target: {value}}) => setSearch(value)}/>
                <Components.Count options={options}>   
                    {count}
                </Components.Count> 
                <Components.GridHeader fields={model.fields} 
                            sort={cursorCalls.sort} 
                            onSort={sort => setCursorCalls({...cursorCalls,
                                sort: [{[sort]: cursorCalls.sort[0][sort] === 1 ? -1 : 1}]
                            })}/>

                {records && <Components.Grid 
                                options={options}
                                skip={skip}
                                count={count}
                                records={records}
                                gridHeight={500} 
                                rowHeight={50} 
                                onScroll={(skip, limit) =>  {
                                    limit = undefined;
                                    model.find(query, {...cursorCalls, limit: [limit], skip: [skip]}).then(records => Promise.all(records)).then(records => setRecords(records)).then(() =>
                                    setCursorCalls({...cursorCalls, limit: [limit], skip: [skip]}))
                                }}/> }
        </>
    )
}

const ObjectShortView = ({children}) => {
    const [record, setRecord] = useState(children)
    if ('then' in children){
        console.log('load')
        children.then(child => setRecord({...child}))
    }
    if (children._id){
        return (
        <div className="ObjectShortView">
            {record.name || record.key || record._id}
        </div>
        )
    }
    else {
        return (
            <pre>
            {JSON.stringify(record, null, 4)}
            </pre>
        )
    }
}


const defaultAdminOptions = 
    {
        view: {
            formatters:{
                ID: ({children}) => <b>{children && children.slice(-6).toUpperCase()}</b>,
                String: ({children}) => <>{children && children.length > 100 ? children.slice(0,100) + '...' : children}</>,
                Object: ObjectShortView,
                Array: ({children}) => <>{children.map(child => <ObjectShortView children={child} />)}</>
            },
            fields:{
                createdAt: ({children}) => <>{new Date(+children).toISOString()}</> ,
                url: ({children}) => <a href={children}>{children}</a> 
            },
            models: {

            }
        },

        edit: {
            formatters:{
                ID: ({children}) => <b>{children && children.slice(-6).toUpperCase()}</b>,
                String: ({children, ...props}) => <textarea value={children} {...props}/>,
                Int: ({children, ...props}) => <input type='number' value={children} {...props}/>,
                Float: ({children, ...props}) => <input type='number' value={children} {...props}/>,
                Object: ObjectShortView,
                Array: ({children}) => <>{children.map(child => <ObjectShortView children={child} />)}</>
            },
            fields:{
                createdAt: ({children}) => <input value={new Date(+children).toISOString()} />,
                url: ({children}) => <a href={children}>{children}</a> 
            },
            models: {

            }
        }
    }


const Admin = ({models, components:{ModelList:ML=ModelList, Search:S=Search, ModelListItem:MLI=ModelListItem, ModelView:MV=ModelView, Count:C=Count, GridHeader:GH=GridHeader, Grid:G=VirtualScroll}={},
                options=defaultAdminOptions
                }) => {
    const [selected, setSelected] = useState()

    const mergedOptions = {...defaultAdminOptions, ...options}


    return (
        <>
            <ML Item={MLI} models={models} onChange={(name) => setSelected(name)} selected={selected}/>
            <content>
                {selected && <MV options={mergedOptions} model={models[selected]} components={{Search: S, Count: C, GridHeader:GH, Grid:G}} /> }
            </content>
        </>
    )
}


function App() {
    let [models, setModels] = useState()
    models || createModels2(gql).then(models => setModels(models))

    const classes = models
    console.log(classes)
    //if (classes && Object.keys(classes).length){
        //classes.Good.find().then(goods => console.log(goods))
        //classes.Image.find().then(images => console.log(images))
    //}

    return (
        <div className="App">
            {models && <Admin models={models} />}
            {/*models && <ModelGrid model={models.Image} rowHeight={50} gridHeight={700} overload={5} query={query} 
                        cellDoubleClick={(name, text, _id) => setQueryText(`{${name}:\`${text}\`}`)}/> */}
        </div>
    );
}

export default App;
